# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Usage: install [-c CLUSTER] [-u USER] [-p PACKAGE] PY_FILES
Installs Python modules in a remote node or cluster.

CLUSTER: Specifies the node or cluster to receive the files. If CLUSTER
matches an osh.remote profile in .oshrc, then the installation is to
that target. If CLUSTER does not match any such profile, then the argument
is interpreted as a hostname. If CLUSTER is omitted, then installation
is to done locally.

USER: The user to connect as. This argument should only be necessary
if CLUSTER specifies a single node. If omitted, then the user is
obtained from the selected osh.remote profile, or if no such profile
has been specified, from the USER environment variable.

PY_FILES: The files to install. The filenames must all end in .py;
they must be Python modules and define classes that extend op or
generator.

PACKAGE: If omitted, files are installed in site-packages. Otherwise,
PACKAGE is converted to a directory path relative to site-packages.
"""

import compiler
import os
import popen2
import sys

import osh.oshobjects
import progtrack

def install(*args):
    return _Install().process_args(*args)

def _scp(user, host, source, target):
    scp_command = 'scp %s %s@%s:%s' % (source, user, host, target)
    read, write, err = popen2.popen3(scp_command)
    err_lines = err.readlines()
    read.close()
    write.close()
    err.close()
    if len(err_lines) > 0:
        raise Exception(' '.join(err_lines))

def _ssh(user, host, command):
    ssh_command = ('ssh %s -T -o StrictHostKeyChecking=no -l %s "%s" ' %
                   (host, user, command))
    read, write, err = popen2.popen3(ssh_command)
    err_lines = err.readlines()
    if len(err_lines) > 0:
        raise Exception(' '.join(err_lines))
    read_lines = read.readlines()
    read.close()
    write.close()
    err.close()
    return read_lines
    

class _Install(osh.oshobjects.RemoteOp):

    _pyfiles = None
    _pycfiles = None
    _ui = None
    _package = None


    # object interface

    def __init__(self):
        osh.oshobjects.RemoteOp.__init__(self, 'c:u:p:', (1, None))

    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        osh.oshobjects.RemoteOp.setup(self,
                                      False # cluster_required = False
                                      )
        self._package = self.args().string_arg('-p')
        self._pyfiles = []
        while self.args().has_next():
            self._pyfiles.append(self.args().next_string())


    # generator interface
    
    def execute(self):
        self.compile()
        if self.hosts():
            self._ui = progtrack.ProgressTrackingUI('install')
            self._ui.add_column('host', 25)
            self._ui.add_column(['find', 'install dir'], 12);
            self._ui.add_column('installed', 12)
            for host in self.hosts():
                self._ui.add_row(host.name())
            self._ui.start()
            try:
                osh.oshobjects.RemoteOp.execute(self)
            finally:
                self._ui.stop()
        else:
            self.install_local()


    # RemoteOp interface

    def action(self, host):
        try:
            # Find install directory
            stage = 1
            version_command = ""
            output = _ssh(self.user(),
                          host.address(),
                          "python -c 'import sys; print sys.version_info[0:2]'")
            package_dir = self.package_dir()
            install_dir = self.install_dir(eval(output[0])) + '/' + package_dir
            if package_dir != '':
                _ssh(self.user(),
                     host.address(),
                     'mkdir -p %s' % install_dir)
                     
            self._ui.ok(host.name(), stage)
            # Copy files
            stage = 2
            _scp(self.user(), host.address(), self.files(), install_dir)
            self._ui.ok(host.name(), stage)
        except:
            (exc_type, exc_value, exc_traceback) = sys.exc_info()
            message = str(exc_value).strip()
            self._ui.error(host.name(), stage, message)


    # For use by this class
    
    def install_local(self):
        install_dir = self.install_dir(sys.version_info[0:2])
        os.system('cp -p %s %s' % (self.files(), install_dir))

    def install_dir(self, version_info):
        return '/usr/lib/python%s.%s/site-packages' % version_info

    def files(self):
        return ' '.join(self._pyfiles + self._pycfiles)

    def compile(self):
        self._pycfiles = []
        for file in self._pyfiles:
            compiler.compileFile(file)
            self._pycfiles.append(file + 'c')

    def package_dir(self):
        if self._package:
            path = self._package.replace('.', '/')
        else:
            path = ''
        return path
