# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Usage: remote @CLUSTER [-u USER] '[' COMMAND ']'
"""

import osh.oshapiparser
import osh.oshargs
import osh.oshobjects
import osh.oshprocess
import osh.osherror

REMOTE_EXECUTABLE = 'remoteosh'

def _dump(stream, object):
    stream.dump(object)

def _consume_remote_stdout(consumer, host, object):
    label = object[0]
    output = object[1]
    if label == 'e':
        problem = output[0]
        if isinstance(problem, osh.osherror.OshError):
            exception = problem.recreate_exception()
        else:
            exception = Exception(problem)
        consumer.send_error(exception)
    else:
        consumer.send((host,) + tuple(output), label)

def _consume_remote_stderr(consumer, host, line):
    # UGLY HACK: remoteosh can occasionally return "[Errno 9] Bad file descriptor" on stderr.
    # I think this is because of io to a process stream whose process has completed.
    # I haven't had luck in tracking this down and fixing the problem for real, so
    # this is a grotesque workaround.
    if '[Errno 9] Bad file descriptor' not in line:
        consumer.send((host, line), osh.oshobjects.STREAM_ERROR)

def remote(*args):
    op = _Remote()
    if len(args) > 0:
        # First arg should specify a cluster
        new_args = [osh.oshargs.Option('-c', args[0])]
        args = args[1:]
        # Second arg should be pipeline to execute remotely
        if len(args) > 0:
            remote_command = args[0]
            if isinstance(remote_command, osh.oshobjects.Op):
                remote_command = [remote_command]
            remote_pipeline = osh.oshapiparser._sequence_op(remote_command)
            op.set_pipeline(remote_pipeline)
            args = args[1:]
        else:
            op.usage()
        args = tuple(new_args) + args
    return op.process_args(*args)

class _Remote(osh.oshobjects.RemoteOp):

    # state

    _pipeline = None


    # object interface
    
    def __init__(self):
        osh.oshobjects.RemoteOp.__init__(self, 'c:u:', (0, 0))
        

    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        osh.oshobjects.RemoteOp.setup(self)


    # generator interface

    def execute(self):
        osh.oshobjects.RemoteOp.execute(self)
        self.send_complete()


    # RemoteOp interface

    def action(self, host):
        process = osh.oshprocess.OshProcess(
            self._remote_command(host.address(), self.user(), host.db_profile()),
            osh.oshprocess.ObjectInputProvider(lambda stream, object:
                                               _dump(stream, object),
                                               [osh.oshobjects.verbosity, self._pipeline]),
            osh.oshprocess.ObjectOutputConsumer(lambda object: _consume_remote_stdout(self,
                                                                                      host.name(),
                                                                                      object)),
            osh.oshprocess.LineOutputConsumer(lambda line: _consume_remote_stderr(self,
                                                                                  host.name(),
                                                                                  line)))
        process.run()
        if process.terminating_exception():
            self.set_terminating_exception(process.terminating_exception())


    # remote compile-time interface

    def set_pipeline(self, pipeline):
        self._pipeline = pipeline
        return self


    # for use by this class

    def _remote_command(self, host, user, db_profile):
        if db_profile:
            remote_command = '"%s %s"' % (REMOTE_EXECUTABLE, db_profile)
        else:
            remote_command = REMOTE_EXECUTABLE
        ssh_command = 'ssh %s -l %s %s' % (host, user, remote_command)
        return ssh_command
