# osh
# Copyright (C) 2005 Jack Orenstein <jao@geophile.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""Usage: window PREDICATE
       window -o WINDOW_SIZE
       window -d WINDOW_SIZE
    
Objects arriving from the input stream are formed into tuples.

The objects are grouped in one of two mechanisms.

1) A new group is started on the first input object, and for any
subsequent object for which PREDICATE returns true. For example, if
the input stream contains the integers 1, 2, 3, ..., then

    window 'n: n % 3 == 2'

yields

    ((1,),)
    ((2,), (3,), (4,))
    ((5,), (6,), (7,))
    ((8,), (9,), (10,))
    ...

I.e., a new tuple is started for each integer n such that n % 3 = 2.

2) Number of objects: -o and -d specifies groups of size
WINDOW_SIZE. 
    
-o specifies 'overlapping' windows, in which each input object begins
a new list containing WINDOW_SIZE items.
    
-d specifies 'disjoint' windows, in which each input object appears in
only one window. A new window is started every WINDOW_SIZE
objects. The last window may be padded with (None,) to ensure that it
has WINDOW_SIZE elements.
    
Example: 
    
    osh gen 10 ^ window -d 3 $
    
yields these tuples:
    
    ((0,), (1,), (2,))
    ((3,), (4,), (5,))
    ((6,), (7,), (8,))
    ((9,), (None,), (None,))
    
Example:

    osh gen 10 ^ window -o 3 $
    
yields these tuples:
    
    ((0,), (1,), (2,))
    ((1,), (2,), (3,))
    ((2,), (3,), (4,))
    ((3,), (4,), (5,))
    ((4,), (5,), (6,))
    ((5,), (6,), (7,))
    ((6,), (7,), (8,))
    ((7,), (8,), (9,))
    ((8,), (9,), (None,))
    ((9,), (None,), (None,))
"""

import types

import osh.oshargs
import osh.oshfunction
import osh.oshobjects

def window(*args):
    return _Window().process_args(*args)

def overlap(n):
    return osh.oshargs.Option('-o', n)

def disjoint(n):
    return osh.oshargs.Option('-d', n)

class _Window(osh.oshobjects.Op):

    _window_generator = None


    # object interface

    def __init__(self):
        osh.oshobjects.Op.__init__(self, 'o:d:', (0, 1))


    # BaseOp interface
    
    def doc(self):
        return __doc__

    def setup(self):
        args = self.args()
        overlap = args.int_arg('-o')
        disjoint = args.int_arg('-d')
        if overlap is not None and disjoint is not None:
            self.usage()
        elif overlap is not None:
            self._window_generator = _OverlapWindow(overlap)
        elif disjoint is not None:
            self._window_generator = _DisjointWindow(disjoint)
        else:
            predicate = args.next_function()
            self._window_generator = _PredicateWindow(predicate)

    def receive(self, object):
        self._window_generator.process(object, self)

    def receive_complete(self):
        self._window_generator.finish(self)

class _OverlapWindow:

    _window_size = None
    _window = None

    def __init__(self, window_size):
        self._window_size = window_size
        self._window = []

    def process(self, object, op):
        if len(self._window) == self._window_size:
            self._window = self._window[1:]
        self._window.append(object)
        if len(self._window) == self._window_size:
            op.send(self._window)

    def finish(self, op):
        padding = (None,)
        if len(self._window) < self._window_size:
            while len(self._window) < self._window_size:
                self._window.append(padding)
            op.send(self._window)
        else:
            for i in xrange(self._window_size - 1):
                self._window = self._window[1:]
                self._window.append(padding)
                op.send(self._window)

class _DisjointWindow:

    _window_size = None
    _window = None

    def __init__(self, window_size):
        self._window_size = window_size
        self._window = []

    def process(self, object, op):
        self._window.append(object)
        if len(self._window) == self._window_size:
            op.send(self._window)
            self._window = []

    def finish(self, op):
        if len(self._window) == 0:
            # Last window was complete
            pass
        else:
            padding = (None,)
            while len(self._window) < self._window_size:
                self._window.append(padding)
            op.send(self._window)

class _PredicateWindow:
    
    _predicate = None
    _window = None

    def __init__(self, predicate):
        self._predicate = predicate
        self._window = []

    def process(self, object, op):
        if self._predicate(object):
            if len(self._window) > 0:
                op.send(self._window)
                self._window = []
        self._window.append(object)

    def finish(self, op):
        op.send(self._window)
